
import enum
from math import ceil
from pathlib import Path
import warnings

import numpy as np

import _libjpeg


LIBJPEG_ERROR_CODES = {
    -1024 : "A parameter for a function was out of range",
    -1025 : "Stream run out of data",
    -1026 : "A code block run out of data",
    -1027 : "Tried to perform an unputc or or an unget on an empty stream",
    -1028 : "Some parameter run out of range",
    -1029 : "The requested operation does not apply",
    -1030 : "Tried to create an already existing object",
    -1031 : "Tried to access a non-existing object",
    -1032 : "A non-optional parameter was left out",
    -1033 : "Forgot to delay a 0xFF",
    -1034 : (
        "Internal error: the requested operation is not available"
    ),
    -1035 : (
        "Internal error: an item computed on a former pass does not "
        "coincide with the same item on a later pass"
    ),
    -1036 : "The stream passed in is no valid jpeg stream",
    -1037 : (
        "A unique marker turned up more than once. The input stream is "
        "most likely corrupt"
    ),
    -1038 : "A misplaced marker segment was found",
    -1040 : (
        "The specified parameters are valid, but are not supported by "
        "the selected profile. Either use a higher profile, or use "
        "simpler parameters (encoder only). "
    ),
    -1041 : (
        "Internal error: the worker thread that was currently active had "
        "to terminate unexpectedly"
    ),
    -1042 : (
        "The encoder tried to emit a symbol for which no Huffman code "
        "was defined. This happens if the standard Huffman table is used "
        "for an alphabet for which it was not defined. The reaction "
        "to this exception should be to create a custom huffman table "
        "instead"
    ),
    -2046 : "Failed to construct the JPEG object",
}


def decode(stream, colour_transform=0, reshape=True):
    """Return the decoded JPEG data from `arr` as a :class:`numpy.ndarray`.

    .. versionchanged:: 1.2

        `stream` can now also be :class:`str` or :class:`pathlib.Path`

    Parameters
    ----------
    stream : str, Path, numpy.ndarray or bytes
        The path to the JPEG file, a 1D array of ``np.uint8``, or a Python
        :class:`bytes` object containing the raw encoded JPEG image.
    colour_transform : int, optional
        The colour transform used, one of:
        | ``0`` : No transform applied (default)
        | ``1`` : RGB to YCbCr
        | ``2`` : JPEG-LS pseudo RCT or RCT
        | ``3`` : Freeform
    reshape : bool, optional
        Reshape and re-view the output array so it matches the image data
        (default), otherwise return a 1D array of ``np.uint8``.

    Returns
    -------
    numpy.ndarray
        An array of containing the decoded image data.

    Raises
    ------
    RuntimeError
        If the decoding failed.
    """
    if isinstance(stream, (str, Path)):
        with open(stream, "rb") as f:
            stream = f.read()

    if isinstance(stream, bytes):
        arr = np.frombuffer(stream, 'uint8')
    else:
        arr = stream

    status, out, params = _libjpeg.decode(arr, colour_transform)
    status = status.decode("utf-8")
    code, msg = status.split("::::")
    code = int(code)

    if code == 0 and reshape is True:
        bpp = ceil(params["precision"] / 8)
        if bpp == 2:
            out = out.view('uint16')

        shape = [params['rows'], params['columns']]
        if params["nr_components"] > 1:
            shape.append(params["nr_components"])

        return out.reshape(*shape)
    elif code == 0 and reshape is False:
        return out

    if code in LIBJPEG_ERROR_CODES:
        raise RuntimeError(
            "libjpeg error code '{}' returned from Decode(): {} - {}"
            .format(code, LIBJPEG_ERROR_CODES[code], msg)
        )

    raise RuntimeError(
        "Unknown error code '{}' returned from Decode(): {}"
        .format(code, msg)
    )


def decode_pixel_data(arr, ds = None, **kwargs):
    """Return the decoded JPEG data from `arr` as a :class:`numpy.ndarray`.

    Intended for use with *pydicom* ``Dataset`` objects.

    Parameters
    ----------
    arr : numpy.ndarray or bytes
        A 1D array of ``np.uint8``, or a Python :class:`bytes` object
        containing the encoded JPEG image.
    ds : pydicom.dataset.Dataset, optional
        A :class:`~pydicom.dataset.Dataset` containing the group ``0x0028``
        elements corresponding to the *Pixel Data*. Must contain a
        (0028,0004) *Photometric Interpretation* element with the colour
        space of the pixel data, one of ``'MONOCHROME1'``, ``'MONOCHROME2'``,
        ``'RGB'``, ``'YBR_FULL'``, ``'YBR_FULL_422'``.
    **kwargs

    Returns
    -------
    numpy.ndarray
        A 1D array of ``numpy.uint8`` containing the decoded image data.

    Raises
    ------
    RuntimeError
        If the decoding failed.
    """
    colours = {
        'MONOCHROME1': 0,
        'MONOCHROME2' : 0,
        'RGB' : 1,
        'YBR_FULL' : 0,
        'YBR_FULL_422' : 0,
    }

    photometric_interpretation = kwargs.get("photometric_interpretation")
    pi = ds.get("PhotometricInterpretation", photometric_interpretation)
    if not pi:
        raise ValueError(
            "The (0028,0004) Photometric Interpretation element is missing "
            "from the dataset"
        )

    try:
        transform = colours[pi]
    except KeyError:
        warnings.warn(
            f"Unsupported (0028,0004) Photometric Interpretation '{pi}', no "
            "colour transformation will be applied"
        )
        transform = 0

    return decode(arr, transform, reshape=False)


def get_parameters(stream):
    """Return a :class:`dict` containing JPEG image parameters.

    .. versionchanged:: 1.2

        `stream` can now also be :class:`str` or :class:`pathlib.Path`

    Parameters
    ----------
    stream : str, Path, numpy.ndarray or bytes
        The path to the JPEG file, a 1D array of ``np.uint8``, or a Python
        :class:`bytes` object containing the raw encoded JPEG image.

    Returns
    -------
    dict
        A :class:`dict` containing JPEG image parameters with keys including
        ``'rows'``, ``'columns'``, ``"nr_components"`` and
        ``"precision"``.

    Raises
    ------
    RuntimeError
        If reading the encoded JPEG data failed.
    """
    if isinstance(stream, (str, Path)):
        with open(stream, "rb") as f:
            stream = f.read()

    if isinstance(stream, bytes):
        arr = np.frombuffer(stream, 'uint8')
    else:
        arr = stream

    status, params = _libjpeg.get_parameters(arr)
    status = status.decode("utf-8")
    code, msg = status.split("::::")
    code = int(code)

    if code == 0:
        return params

    if code in LIBJPEG_ERROR_CODES:
        raise RuntimeError(
            "libjpeg error code '{}' returned from GetJPEGParameters(): "
            "{} - {}".format(code, LIBJPEG_ERROR_CODES[code], msg)
        )

    raise RuntimeError(
        "Unknown error code '{}' returned from GetJPEGParameters(): {}"
        .format(status, msg)
    )


def reconstruct(fin, fout, colourspace=1, falpha=None, upsample=True):
    """Simple wrapper for the libjpeg ``cmd/reconstruct::Reconstruct()``
    function.

    Parameters
    ----------
    fin : bytes
        The path to the JPEG file to be decoded.
    fout : bytes
        The path to the decoded PPM or PGM (if `falpha` is ``True``) file(s).
    colourspace : int, optional
        The colourspace transform to apply.
        | ``0`` : ``JPGFLAG_MATRIX_COLORTRANSFORMATION_NONE``  (``-c`` flag)
        | ``1`` : ``JPGFLAG_MATRIX_COLORTRANSFORMATION_YCBCR`` (default)
        | ``2`` : ``JPGFLAG_MATRIX_COLORTRANSFORMATION_LSRCT`` (``-cls`` flag)
        | ``2`` : ``JPGFLAG_MATRIX_COLORTRANSFORMATION_RCT``
        | ``3`` : ``JPGFLAG_MATRIX_COLORTRANSFORMATION_FREEFORM``
        See `here<https://github.com/thorfdbg/libjpeg/blob/87636f3b26b41b85b2fb7355c589a8c456ef808c/interface/parameters.hpp#L381>`_
        for more information.
    falpha : bytes, optional
        The path where any decoded alpha channel data will be written (as a
        PGM file), otherwise ``None`` (default) to not write alpha channel
        data. Equivalent to the ``-al file`` flag.
    upsample : bool, optional
        ``True`` (default) to disable automatic upsampling, equivalent to
        the ``-U`` flag.
    """
    if isinstance(fin, (str, Path)):
        fin = str(fin)
        fin = bytes(fin, 'utf-8')

    if isinstance(fout, (str, Path)):
        fout = str(fout)
        fout = bytes(fout, 'utf-8')

    if falpha and isinstance(falpha, (str, Path)):
        falpha = str(falpha)
        falpha = bytes(falpha, 'utf-8')

    _libjpeg.reconstruct(fin, fout, colourspace, falpha, upsample)



"""Tests for the pylibjpeg pixel data handler."""

from io import BytesIO
import os
import pytest
import warnings

import numpy as np

try:
    import pydicom
    import pydicom.config
    from pydicom.pixel_data_handlers.util import convert_color_space
    from pydicom.encaps import defragment_data, generate_pixel_data_frame
    HAS_PYDICOM = True
except ImportError:
    HAS_PYDICOM = False

from libjpeg import get_parameters
from libjpeg.data import get_indexed_datasets, JPEG_DIRECTORY


DIR_10918 = os.path.join(JPEG_DIRECTORY, '10918')
DIR_14495 = os.path.join(JPEG_DIRECTORY, '14495')


REF_DCM = {
    '1.2.840.10008.1.2.4.50' : [
        # filename, (rows, columns, samples/px, bits/sample)
        ('JPEGBaseline_1s_1f_u_08_08.dcm', (100, 100, 1, 8)),
        ('SC_rgb_dcmtk_+eb+cy+np.dcm', (100, 100, 3, 8)),
        ('color3d_jpeg_baseline.dcm', (480, 640, 3, 8)),
        ('SC_rgb_dcmtk_+eb+cr.dcm', (100, 100, 3, 8)),
        ('SC_rgb_dcmtk_+eb+cy+n1.dcm', (100, 100, 3, 8)),
        ('SC_rgb_dcmtk_+eb+cy+s4.dcm', (100, 100, 3, 8)),
    ],
    '1.2.840.10008.1.2.4.51' : [
        ('RG2_JPLY_fixed.dcm', (2140, 1760, 1, 12)),
        ('JPEGExtended_1s_1f_u_16_12.dcm', (1024, 256, 1, 12)),
        ('JPEGExtended_3s_1f_u_08_08.dcm', (576, 768, 3, 8)),
    ],
    '1.2.840.10008.1.2.4.57' : [
        ('JPEGLossless_1s_1f_u_16_12.dcm', (1024, 1024, 1, 12)),
    ],
    '1.2.840.10008.1.2.4.70' : [
        ('JPEG-LL.dcm', (1024, 256, 1, 16)),
        ('JPEGLosslessP14SV1_1s_1f_u_08_08.dcm', (768, 1024, 1, 8)),
        ('JPEGLosslessP14SV1_1s_1f_u_16_16.dcm', (535, 800, 1, 16)),
        ('MG1_JPLL.dcm', (4664, 3064, 1, 12)),
        ('RG1_JPLL.dcm', (1955, 1841, 1, 15)),
        ('RG2_JPLL.dcm', (2140, 1760, 1, 10)),
        ('SC_rgb_jpeg_gdcm.dcm', (100, 100, 3, 8)),
    ],
    '1.2.840.10008.1.2.4.80' : [
        ('emri_small_jpeg_ls_lossless.dcm', (64, 64, 1, 12)),
        ('MR_small_jpeg_ls_lossless.dcm', (64, 64, 1, 16)),
        ('RG1_JLSL.dcm', (1955, 1841, 1, 16)),
        ('RG2_JLSL.dcm', (2140, 1760, 1, 10)),
    ],
    '1.2.840.10008.1.2.4.81' : [
        ('CT1_JLSN.dcm', (512, 512, 1, 16)),
        ('MG1_JLSN.dcm', (4664, 3064, 1, 12)),
        ('RG1_JLSN.dcm', (1955, 1841, 1, 15)),
        ('RG2_JLSN.dcm', (2140, 1760, 1, 10)),
    ],
}


def test_get_parameters_bytes():
    """Test get_parameters() using bytes."""
    with open(os.path.join(DIR_10918, 'p1', 'A1.JPG'), 'rb') as fp:
        data = fp.read()

    params = get_parameters(data)

    info = (257, 255, 4, 8)

    assert (info[0], info[1]) == (params['rows'], params['columns'])
    assert info[2] == params["nr_components"]
    assert info[3] == params["precision"]


@pytest.mark.skipif(not HAS_PYDICOM, reason="No pydicom")
def test_non_conformant_raises():
    """Test that a non-conformant JPEG image raises an exception."""
    ds_list = get_indexed_datasets('1.2.840.10008.1.2.4.51')
    # Image has invalid Se value in the SOS marker segment
    item = ds_list['JPEG-lossy.dcm']
    assert 0xC000 == item['Status'][1]
    ds = item['ds']

    nr_frames = ds.get('NumberOfFrames', 1)
    frame = next(generate_pixel_data_frame(ds.PixelData, nr_frames))

    msg = (
        r"libjpeg error code '-1038' returned from GetJPEGParameters\(\): A "
        r"misplaced marker segment was found - scan start must be zero "
        r"and scan stop must be 63 for the sequential operating modes"
    )
    with pytest.raises(RuntimeError, match=msg):
        get_parameters(frame)


@pytest.mark.skip("Not sure how to trigger exception")
def test_unknown_error_raised():
    """Test that an unknown error is handled properly."""
    #get_parameters(b'')
    pass


@pytest.mark.skipif(not HAS_PYDICOM, reason="No pydicom")
class TestGetParametersDCM(object):
    """Tests for get_parameters() using DICOM datasets."""
    def generate_frames(self, ds):
        """Return a generator object with the dataset's pixel data frames."""
        nr_frames = ds.get('NumberOfFrames', 1)
        return generate_pixel_data_frame(ds.PixelData, nr_frames)

    @pytest.mark.parametrize("fname, info", REF_DCM['1.2.840.10008.1.2.4.50'])
    def test_baseline(self, fname, info):
        """Test get_parameters() for the baseline datasets."""
        #info: (rows, columns, spp, bps)
        index = get_indexed_datasets('1.2.840.10008.1.2.4.50')
        ds = index[fname]['ds']

        frame = next(self.generate_frames(ds))
        params = get_parameters(np.frombuffer(frame, 'uint8'))

        assert (info[0], info[1]) == (params['rows'], params['columns'])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_DCM['1.2.840.10008.1.2.4.51'])
    def test_extended(self, fname, info):
        """Test get_parameters() for the baseline datasets."""
        #info: (rows, columns, spp, bps)
        index = get_indexed_datasets('1.2.840.10008.1.2.4.51')
        ds = index[fname]['ds']

        frame = next(self.generate_frames(ds))
        params = get_parameters(np.frombuffer(frame, 'uint8'))

        assert (info[0], info[1]) == (params['rows'], params['columns'])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_DCM['1.2.840.10008.1.2.4.57'])
    def test_lossless(self, fname, info):
        """Test get_parameters() for the lossless datasets."""
        #info: (rows, columns, spp, bps)
        index = get_indexed_datasets('1.2.840.10008.1.2.4.57')
        ds = index[fname]['ds']

        frame = next(self.generate_frames(ds))
        params = get_parameters(np.frombuffer(frame, 'uint8'))

        assert (info[0], info[1]) == (params['rows'], params['columns'])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_DCM['1.2.840.10008.1.2.4.70'])
    def test_lossless_sv1(self, fname, info):
        """Test get_parameters() for the lossless SV1 datasets."""
        #info: (rows, columns, spp, bps)
        index = get_indexed_datasets('1.2.840.10008.1.2.4.70')
        ds = index[fname]['ds']

        frame = next(self.generate_frames(ds))
        params = get_parameters(np.frombuffer(frame, 'uint8'))

        assert (info[0], info[1]) == (params['rows'], params['columns'])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_DCM['1.2.840.10008.1.2.4.80'])
    def test_extended(self, fname, info):
        """Test get_parameters() for the LS lossless datasets."""
        #info: (rows, columns, spp, bps)
        index = get_indexed_datasets('1.2.840.10008.1.2.4.80')
        ds = index[fname]['ds']

        frame = next(self.generate_frames(ds))
        params = get_parameters(np.frombuffer(frame, 'uint8'))

        assert (info[0], info[1]) == (params['rows'], params['columns'])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_DCM['1.2.840.10008.1.2.4.81'])
    def test_extended(self, fname, info):
        """Test get_parameters() for the LS lossy datasets."""
        #info: (rows, columns, spp, bps)
        index = get_indexed_datasets('1.2.840.10008.1.2.4.81')
        ds = index[fname]['ds']

        frame = next(self.generate_frames(ds))
        params = get_parameters(np.frombuffer(frame, 'uint8'))

        assert (info[0], info[1]) == (params['rows'], params['columns'])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]


REF_JPG = {
    '10918' : {
        'p1' : [
            ('A1.JPG', (257, 255, 4, 8)),
            #('B1.JPG', (1, 1, 1, 1)),
            #('B2.JPG', (1, 1, 1, 1)), missing DHT marker (its in B1.JPG)
        ],
        'p2' : [
            ('A1.JPG', (257, 255, 4, 8)),
            #('B1.JPG', ()),
            #('B2.JPG', ()), missing DHT (in B1)
            ('C1.JPG', (257, 255, 4, 8)),
            ('C2.JPG', (257, 255, 4, 8)),
        ],
        'p4' : [
            ('A1.JPG', (257, 255, 4, 8)),
            #('B1.JPG', ()),
            #('B2.JPG', ()),
            ('C1.JPG', (257, 255, 4, 8)),
            ('C2.JPG', (257, 255, 4, 8)),
            ('E1.JPG', (257, 255, 4, 12)),
            ('E2.JPG', (257, 255, 4, 12)),
        ],
        'p14' : [
            ('O1.JPG', (257, 255, 4, 8)),
            ('O2.JPG', (257, 255, 4, 16)),
        ],
    },
    '14495' : {
        'JLS' : [
            ('T8C0E0.JLS', (256, 256, 3, 8)),
            ('T8C1E0.JLS', (256, 256, 3, 8)),
            ('T8C2E0.JLS', (256, 256, 3, 8)),
            ('T8NDE0.JLS', (128, 128, 1, 8)),
            ('T8SSE0.JLS', (256, 256, 3, 8)),
            ('T16E0.JLS', (256, 256, 1, 12)),
        ],
        'JNL' : [
            ('T8C0E3.JLS', (256, 256, 3, 8)),
            ('T8C1E3.JLS', (256, 256, 3, 8)),
            ('T8C2E3.JLS', (256, 256, 3, 8)),
            ('T8NDE3.JLS', (128, 128, 1, 8)),
            ('T8SSE3.JLS', (256, 256, 3, 8)),
            ('T16E3.JLS', (256, 256, 1, 12)),
        ],
    },
    '15444' : {},
}


class TestGetParametersJPG(object):
    """Tests for get_parameters() using JPEG compliance data."""
    @pytest.mark.parametrize("fname, info", REF_JPG['10918']['p1'])
    def test_baseline(self, fname, info):
        """Test get_parameters() for the baseline compliance images."""
        #info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_10918, 'p1', fname), 'rb') as fp:
            data = fp.read()

        params = get_parameters(np.frombuffer(data, 'uint8'))

        assert (info[0], info[1]) == (params['rows'], params['columns'])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_JPG['10918']['p2'])
    def test_extended_p2(self, fname, info):
        """Test get_parameters() for the extended p2 compliance images."""
        #info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_10918, 'p2', fname), 'rb') as fp:
            data = fp.read()

        params = get_parameters(np.frombuffer(data, 'uint8'))

        assert (info[0], info[1]) == (params['rows'], params['columns'])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_JPG['10918']['p4'])
    def test_extended_p4(self, fname, info):
        """Test get_parameters() for the extended p4 compliance images."""
        #info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_10918, 'p4', fname), 'rb') as fp:
            data = fp.read()

        params = get_parameters(np.frombuffer(data, 'uint8'))

        assert (info[0], info[1]) == (params['rows'], params['columns'])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_JPG['10918']['p14'])
    def test_lossless_p14(self, fname, info):
        """Test get_parameters() for the extended p14 compliance images."""
        #info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_10918, 'p14', fname), 'rb') as fp:
            data = fp.read()

        params = get_parameters(np.frombuffer(data, 'uint8'))

        assert (info[0], info[1]) == (params['rows'], params['columns'])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]

    @pytest.mark.parametrize("fname, info", REF_JPG['14495']['JLS'])
    def test_jls(self, fname, info):
        """Test get_parameters() for the JPEG-LS compliance images."""
        #info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_14495, 'JLS', fname), 'rb') as fp:
            data = fp.read()

        params = get_parameters(np.frombuffer(data, 'uint8'))

        assert (info[0], info[1]) == (params['rows'], params['columns'])
        assert info[2] == params["nr_components"]
        assert info[3] == params["precision"]



"""Tests for pylibjpeg.decode()."""

from io import BytesIO
import os
from pathlib import Path
import pytest
import warnings

import numpy as np

try:
    import pydicom
    from pydicom.encaps import generate_pixel_data_frame
    from pydicom.pixel_data_handlers.util import reshape_pixel_array
    HAS_PYDICOM = True
except ImportError:
    HAS_PYDICOM = False

from libjpeg import decode, decode_pixel_data
from libjpeg.data import get_indexed_datasets, JPEG_DIRECTORY


DIR_10918 = JPEG_DIRECTORY / '10918'
DIR_14495 = JPEG_DIRECTORY / '14495'


REF_DCM = {
    '1.2.840.10008.1.2.4.50' : [
        # filename, (rows, columns, samples/px, bits/sample)
        ('JPEGBaseline_1s_1f_u_08_08.dcm', (100, 100, 1, 8)),
        ('SC_rgb_dcmtk_+eb+cy+np.dcm', (100, 100, 3, 8)),
        ('color3d_jpeg_baseline.dcm', (480, 640, 3, 8)),
        ('SC_rgb_dcmtk_+eb+cr.dcm', (100, 100, 3, 8)),
        ('SC_rgb_dcmtk_+eb+cy+n1.dcm', (100, 100, 3, 8)),
        ('SC_rgb_dcmtk_+eb+cy+s4.dcm', (100, 100, 3, 8)),
    ],
    '1.2.840.10008.1.2.4.51' : [
        ('RG2_JPLY_fixed.dcm', (2140, 1760, 1, 12)),
        ('JPEGExtended_1s_1f_u_16_12.dcm', (1024, 256, 1, 12)),
        ('JPEGExtended_3s_1f_u_08_08.dcm', (576, 768, 3, 8)),
    ],
    '1.2.840.10008.1.2.4.57' : [
        ('JPEGLossless_1s_1f_u_16_12.dcm', (1024, 1024, 1, 12)),
    ],
    '1.2.840.10008.1.2.4.70' : [
        ('JPEG-LL.dcm', (1024, 256, 1, 16)),
        ('JPEGLosslessP14SV1_1s_1f_u_08_08.dcm', (768, 1024, 1, 8)),
        ('JPEGLosslessP14SV1_1s_1f_u_16_16.dcm', (535, 800, 1, 16)),
        ('MG1_JPLL.dcm', (4664, 3064, 1, 12)),
        ('RG1_JPLL.dcm', (1955, 1841, 1, 15)),
        ('RG2_JPLL.dcm', (2140, 1760, 1, 10)),
        ('SC_rgb_jpeg_gdcm.dcm', (100, 100, 3, 8)),
    ],
    '1.2.840.10008.1.2.4.80' : [
        ('emri_small_jpeg_ls_lossless.dcm', (64, 64, 1, 12)),
        ('MR_small_jpeg_ls_lossless.dcm', (64, 64, 1, 16)),
        ('RG1_JLSL.dcm', (1955, 1841, 1, 16)),
        ('RG2_JLSL.dcm', (2140, 1760, 1, 10)),
    ],
    '1.2.840.10008.1.2.4.81' : [
        ('CT1_JLSN.dcm', (512, 512, 1, 16)),
        ('MG1_JLSN.dcm', (4664, 3064, 1, 12)),
        ('RG1_JLSN.dcm', (1955, 1841, 1, 15)),
        ('RG2_JLSN.dcm', (2140, 1760, 1, 10)),
    ],
}


# TODO: convert using straight JPG data
@pytest.mark.skipif(not HAS_PYDICOM, reason="No pydicom")
def test_decode_bytes():
    """Test decode using bytes."""
    index = get_indexed_datasets('1.2.840.10008.1.2.4.50')
    ds = index['JPEGBaseline_1s_1f_u_08_08.dcm']['ds']
    nr_frames = ds.get('NumberOfFrames', 1)
    frame = next(generate_pixel_data_frame(ds.PixelData, nr_frames))
    arr = decode(frame)
    assert arr.flags.writeable
    assert 'uint8' == arr.dtype
    assert (ds.Rows, ds.Columns) == arr.shape

    # Reference values from GDCM handler
    assert  76 == arr[ 5, 50]
    assert 167 == arr[15, 50]
    assert 149 == arr[25, 50]
    assert 203 == arr[35, 50]
    assert  29 == arr[45, 50]
    assert 142 == arr[55, 50]
    assert   1 == arr[65, 50]
    assert  64 == arr[75, 50]
    assert 192 == arr[85, 50]
    assert 255 == arr[95, 50]


# TODO: convert to using straight JPG data
@pytest.mark.skipif(not HAS_PYDICOM, reason="No pydicom")
def test_invalid_colourspace_warns():
    """Test that using an unknown colourspace gives a warning."""
    index = get_indexed_datasets('1.2.840.10008.1.2.4.50')
    ds = index['JPEGBaseline_1s_1f_u_08_08.dcm']['ds']
    nr_frames = ds.get('NumberOfFrames', 1)
    frame = next(generate_pixel_data_frame(ds.PixelData, nr_frames))
    msg = r"no colour transformation will be applied"
    ds.PhotometricInterpretation = 'ANY'
    with pytest.warns(UserWarning, match=msg):
        arr = decode_pixel_data(np.frombuffer(frame, 'uint8'), ds)

    arr = reshape_pixel_array(ds, arr)

    assert arr.flags.writeable
    assert 'uint8' == arr.dtype
    assert (ds.Rows, ds.Columns) == arr.shape

    # Reference values from GDCM handler
    assert  76 == arr[ 5, 50]
    assert 167 == arr[15, 50]
    assert 149 == arr[25, 50]
    assert 203 == arr[35, 50]
    assert  29 == arr[45, 50]
    assert 142 == arr[55, 50]
    assert   1 == arr[65, 50]
    assert  64 == arr[75, 50]
    assert 192 == arr[85, 50]
    assert 255 == arr[95, 50]


@pytest.mark.skipif(not HAS_PYDICOM, reason="No pydicom")
class TestDecodeDCM(object):
    """Tests for get_parameters() using DICOM datasets."""
    def generate_frames(self, ds):
        """Return a generator object with the dataset's pixel data frames."""
        nr_frames = ds.get('NumberOfFrames', 1)
        return generate_pixel_data_frame(ds.PixelData, nr_frames)

    @pytest.mark.parametrize("fname, info", REF_DCM['1.2.840.10008.1.2.4.50'])
    def test_baseline(self, fname, info):
        """Test get_parameters() for the baseline datasets."""
        #info: (rows, columns, spp, bps)
        index = get_indexed_datasets('1.2.840.10008.1.2.4.50')
        ds = index[fname]['ds']

        frame = next(self.generate_frames(ds))
        arr = decode(np.frombuffer(frame, 'uint8'), reshape=True)

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == 'uint8'
        if 9 <= info[3] <= 16:
            assert arr.dtype == 'uint16'

    @pytest.mark.parametrize("fname, info", REF_DCM['1.2.840.10008.1.2.4.51'])
    def test_extended(self, fname, info):
        """Test get_parameters() for the baseline datasets."""
        #info: (rows, columns, spp, bps)
        index = get_indexed_datasets('1.2.840.10008.1.2.4.51')
        ds = index[fname]['ds']

        frame = next(self.generate_frames(ds))
        arr = decode(np.frombuffer(frame, 'uint8'), reshape=True)

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == 'uint8'
        if 9 <= info[3] <= 16:
            assert arr.dtype == 'uint16'

    @pytest.mark.parametrize("fname, info", REF_DCM['1.2.840.10008.1.2.4.57'])
    def test_lossless(self, fname, info):
        """Test get_parameters() for the lossless datasets."""
        #info: (rows, columns, spp, bps)
        index = get_indexed_datasets('1.2.840.10008.1.2.4.57')
        ds = index[fname]['ds']

        frame = next(self.generate_frames(ds))
        arr = decode(np.frombuffer(frame, 'uint8'), reshape=True)

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == 'uint8'
        if 9 <= info[3] <= 16:
            assert arr.dtype == 'uint16'

    @pytest.mark.parametrize("fname, info", REF_DCM['1.2.840.10008.1.2.4.70'])
    def test_lossless_sv1(self, fname, info):
        """Test get_parameters() for the lossless SV1 datasets."""
        #info: (rows, columns, spp, bps)
        index = get_indexed_datasets('1.2.840.10008.1.2.4.70')
        ds = index[fname]['ds']

        frame = next(self.generate_frames(ds))
        arr = decode(np.frombuffer(frame, 'uint8'), reshape=True)

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == 'uint8'
        if 9 <= info[3] <= 16:
            assert arr.dtype == 'uint16'

    @pytest.mark.parametrize("fname, info", REF_DCM['1.2.840.10008.1.2.4.80'])
    def test_extended(self, fname, info):
        """Test get_parameters() for the LS lossless datasets."""
        #info: (rows, columns, spp, bps)
        index = get_indexed_datasets('1.2.840.10008.1.2.4.80')
        ds = index[fname]['ds']

        frame = next(self.generate_frames(ds))
        arr = decode(np.frombuffer(frame, 'uint8'), reshape=True)

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == 'uint8'
        if 9 <= info[3] <= 16:
            assert arr.dtype == 'uint16'

    @pytest.mark.parametrize("fname, info", REF_DCM['1.2.840.10008.1.2.4.81'])
    def test_extended(self, fname, info):
        """Test get_parameters() for the LS lossy datasets."""
        #info: (rows, columns, spp, bps)
        index = get_indexed_datasets('1.2.840.10008.1.2.4.81')
        ds = index[fname]['ds']

        frame = next(self.generate_frames(ds))
        arr = decode(np.frombuffer(frame, 'uint8'), reshape=True)

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == 'uint8'
        if 9 <= info[3] <= 16:
            assert arr.dtype == 'uint16'


REF_JPG = {
    '10918' : {
        # Reference values using libjpeg's `jpeg` app + imageJ
        # Very circular but having issues with other programs interpreting
        #   it as CYMK and doing weird stuff
        'p1' : [
            # ([top left pixel], [bottom right pixel])
            (
                'A1.JPG', (257, 255, 4, 8),
                ([138, 76, 239, 216], [155, 217, 191, 115])
            ),
            #('B1.JPG', (1, 1, 1, 1)),
            #('B2.JPG', (1, 1, 1, 1)), missing DHT marker (its in B1.JPG)
        ],
        'p2' : [
            (
                'A1.JPG', (257, 255, 4, 8),
                ([138, 76, 239, 216], [155, 217, 191, 115])
            ),
            #('B1.JPG', ()),
            #('B2.JPG', ()), missing DHT (in B1)
            (
                'C1.JPG', (257, 255, 4, 8),
                ([138, 76, 239, 216], [155, 217, 191, 115])
            ),
            (
                'C2.JPG', (257, 255, 4, 8),
                ([138, 76, 239, 216], [155, 217, 191, 115])
            ),
        ],
        'p4' : [
            (
                'A1.JPG', (257, 255, 4, 8),
                ([138, 76, 239, 216], [155, 217, 191, 115])
            ),
            #('B1.JPG', ()),
            #('B2.JPG', ()),
            (
                'C1.JPG', (257, 255, 4, 8),
                ([138, 76, 239, 216], [155, 217, 191, 115])
            ),
            (
                'C2.JPG', (257, 255, 4, 8),
                ([138, 76, 239, 216], [155, 217, 191, 115])
            ),
            (
                'E1.JPG', (257, 255, 4, 12),
                ([2119, 1183, 3907, 3487], [2502, 3402, 3041, 1872])
            ),
            (
                'E2.JPG', (257, 255, 4, 12),
                ([2119, 1183, 3907, 3487], [2502, 3402, 3041, 1872])
            ),
        ],
        'p14' : [
            (
                'O1.JPG', (257, 255, 4, 8),
                ([132, 76, 245, 218], [156, 211, 191, 116])
            ),
            (
                'O2.JPG', (257, 255, 4, 16),
                ([33792, 19456, 62720, 55808], [39936, 53888, 48768, 29696])
            ),
        ],
    },
    '14495' : {
        'JLS' : [  # Lossless
            (
                # Reference values using GDCM
                'T8C0E0.JLS', (256, 256, 3, 8),
                ([161, 122, 108], [101, 99, 95])
            ),
            (
                # GDCM and CharPyLS fail on this one - both use CharLS
                 # Circular reference
                'T8C1E0.JLS', (256, 256, 3, 8),
                ([161, 122, 108], [101, 99, 95])
            ),
            (
                # CharLS fails
                # Circular reference
                'T8C2E0.JLS', (256, 256, 3, 8),
                ([161, 122, 108], [101, 99, 95])
            ),
            (
                # Reference values using GDCM
                'T8NDE0.JLS', (128, 128, 1, 8),
                (108, 231)
            ),
            (
                # CharLS fails
                # Circular reference
                'T8SSE0.JLS', (256, 256, 3, 8),
                ([161, 122, 108], [101, 171, 231])
            ),
            (
                # Reference values using GDCM
                'T16E0.JLS', (256, 256, 1, 12),
                (1963, 1596)
            ),
        ],
        'JNL' : [  # Near lossless
            (
                # Reference values using GDCM
                'T8C0E3.JLS', (256, 256, 3, 8),
                ([161, 119, 105], [98, 96, 93])
            ),
            (
                # CharLS fails
                # Circular reference
                'T8C1E3.JLS', (256, 256, 3, 8),
                ([161, 119, 105], [101, 100, 97])
            ),
            (
                # CharLS fails
                # Circular reference
                'T8C2E3.JLS', (256, 256, 3, 8),
                ([161, 119, 105], [98, 96, 94])
            ),
            (
                # Reference values using GDCM
                'T8NDE3.JLS', (128, 128, 1, 8),
                (105, 229)
            ),
            (
                # CharLS fails
                # Circular reference
                'T8SSE3.JLS', (256, 256, 3, 8),
                ([161, 119, 105], [102, 169, 234])
            ),
            (
                # Reference values using GDCM
                'T16E3.JLS', (256, 256, 1, 12),
                (1960, 1593)
            ),
        ],
    },
    '15444' : {},
}


class TestDecodeJPG(object):
    """Tests for get_parameters() using JPEG compliance data."""
    @pytest.mark.parametrize("fname, info, ref", REF_JPG['10918']['p1'])
    def test_baseline(self, fname, info, ref):
        """Test decoding the baseline compliance images."""
        #info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_10918, 'p1', fname), 'rb') as fp:
            data = fp.read()

        arr = decode(np.frombuffer(data, 'uint8'), reshape=True)
        assert arr.flags.writeable

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        # Process 1 is always 8-bit
        assert arr.dtype == 'uint8'

        assert ref[0] == arr[0, 0, :].tolist()
        assert ref[1] == arr[-1, -1, :].tolist()

    @pytest.mark.parametrize("fname, info, ref", REF_JPG['10918']['p2'])
    def test_extended_p2(self, fname, info, ref):
        """Test decoding the extended p2 compliance images."""
        #info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_10918, 'p2', fname), 'rb') as fp:
            data = fp.read()

        arr = decode(np.frombuffer(data, 'uint8'), reshape=True)
        assert arr.flags.writeable

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        # Process 2 is always 8-bit
        assert arr.dtype == 'uint8'

        assert ref[0] == arr[0, 0, :].tolist()
        assert ref[1] == arr[-1, -1, :].tolist()

    @pytest.mark.parametrize("fname, info, ref", REF_JPG['10918']['p4'])
    def test_extended_p4(self, fname, info, ref):
        """Test decoding the extended p4 compliance images."""
        #info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_10918, 'p4', fname), 'rb') as fp:
            data = fp.read()

        arr = decode(np.frombuffer(data, 'uint8'), reshape=True)
        assert arr.flags.writeable

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == 'uint8'
        if 9 <= info[3] <= 16:
            assert arr.dtype == 'uint16'

        assert ref[0] == arr[0, 0, :].tolist()
        assert ref[1] == arr[-1, -1, :].tolist()

    @pytest.mark.parametrize("fname, info, ref", REF_JPG['10918']['p14'])
    def test_lossless_p14(self, fname, info, ref):
        """Test decoding the extended p14 compliance images."""
        #info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_10918, 'p14', fname), 'rb') as fp:
            data = fp.read()

        arr = decode(np.frombuffer(data, 'uint8'), reshape=True)
        assert arr.flags.writeable

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == 'uint8'
        if 9 <= info[3] <= 16:
            assert arr.dtype == 'uint16'

        assert ref[0] == arr[0, 0, :].tolist()
        assert ref[1] == arr[-1, -1, :].tolist()

    @pytest.mark.parametrize("fname, info, ref", REF_JPG['14495']['JLS'])
    def test_jls(self, fname, info, ref):
        """Test decoding the JPEG-LS lossless compliance images."""
        #info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_14495, 'JLS', fname), 'rb') as fp:
            data = fp.read()

        arr = decode(
            np.frombuffer(data, 'uint8'),
            reshape=True,
        )
        assert arr.flags.writeable

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == 'uint8'
        if 9 <= info[3] <= 16:
            assert arr.dtype == 'uint16'

        if info[2] == 1:
            assert ref[0] == arr[0, 0].tolist()
            assert ref[1] == arr[-1, -1].tolist()
        else:
            assert ref[0] == arr[0, 0, :].tolist()
            assert ref[1] == arr[-1, -1, :].tolist()

    @pytest.mark.parametrize("fname, info, ref", REF_JPG['14495']['JNL'])
    def test_jnl(self, fname, info, ref):
        """Test decoding the JPEG-LS near lossless compliance images."""
        #info: (rows, columns, spp, bps)
        with open(os.path.join(DIR_14495, 'JNL', fname), 'rb') as fp:
            data = fp.read()

        arr = decode(np.frombuffer(data, 'uint8'), reshape=True)
        assert arr.flags.writeable

        if info[2] == 1:
            assert (info[0], info[1]) == arr.shape
        else:
            assert (info[0], info[1], info[2]) == arr.shape

        if 1 <= info[3] <= 8:
            assert arr.dtype == 'uint8'
        if 9 <= info[3] <= 16:
            assert arr.dtype == 'uint16'

        if info[2] == 1:
            assert ref[0] == arr[0, 0].tolist()
            assert ref[1] == arr[-1, -1].tolist()
        else:
            assert ref[0] == arr[0, 0, :].tolist()
            assert ref[1] == arr[-1, -1, :].tolist()

    def test_str(self):
        """Test decoding using a str path."""
        p = DIR_10918 / 'p1' / 'A1.JPG'
        p = os.fspath(p)
        assert isinstance(p, str)
        arr = decode(p)
        assert arr.shape == (257, 255, 4)

    def test_path(self):
        """Test decoding using a Path path."""
        p = DIR_10918 / 'p1' / 'A1.JPG'
        assert isinstance(p, Path)
        arr = decode(p)
        assert arr.shape == (257, 255, 4)



"""Tests for the pylibjpeg pixel data handler."""

from io import BytesIO
import os
import tempfile
from tempfile import NamedTemporaryFile
import pytest

from libjpeg.utils import reconstruct
from libjpeg.data import JPEG_DIRECTORY


DIR_10918 = os.path.join(JPEG_DIRECTORY, '10918')
DIR_14495 = os.path.join(JPEG_DIRECTORY, '14495')


def test_reconstruct_no_alpha():
    """Basic test of reconstruct()."""
    # Windows workarounds
    tempdir = tempfile.TemporaryDirectory()
    inname = os.path.join(tempdir.name, os.urandom(24).hex())
    outname = os.path.join(tempdir.name, os.urandom(24).hex())

    infile = open(inname, 'wb')
    with open(os.path.join(DIR_14495, 'JLS', 'T8C0E0.JLS'), 'rb') as fp:
        infile.write(fp.read())

    infile.close()

    # Output file
    reconstruct(
        inname, outname, colourspace=0, falpha=None, upsample=True
    )
    with open(outname, 'rb') as f:
        data = f.read()

    tempdir.cleanup()

    assert b'P6\n256 256' == data[:10]


@pytest.mark.skip("Needs more understanding")
def test_reconstruct_alpha():
    """Basic test of reconstruct() with pretend alpha data."""
    # Input file
    infile = NamedTemporaryFile('rb+')
    with open(os.path.join(DIR_10918, 'p1', 'A1.JPG'), 'rb') as fp:
        infile.write(fp.read())

    # Output file - CMY
    outfile = NamedTemporaryFile('rb+')
    # Output file - "alpha"
    alphafile = NamedTemporaryFile('rb+')

    reconstruct(
        infile.name, outfile.name, colourspace=0,
        falpha=alphafile.name, upsample=True
    )

    assert b'' == alphafile.read(10)
    assert b'P6\n256 256' == outfile.read(10)



# cython: language_level=3
# distutils: language=c++

from math import ceil

from libcpp cimport bool
from libcpp.string cimport string

import numpy as np
cimport numpy as np


cdef extern from "decode.hpp":
    cdef string Decode(
        char *inArray,
        char *outArray,
        int inLength,
        int outLength,
        int colourspace,
    )

    cdef struct JPEGParameters:
        int marker
        long columns
        long rows
        int samples_per_pixel
        char bits_per_sample

    cdef string GetJPEGParameters(
        char *inArray,
        int inLength,
        JPEGParameters *param,
    )


def decode(np.ndarray[np.uint8_t, ndim=1] input_buffer, colourspace):
    """Return the decoded JPEG data from `input_buffer`.

    Parameters
    ----------
    input_buffer : numpy.ndarray
        A 1D array of ``np.uint8`` containing the raw encoded JPEG image.
    colourspace : int
        | ``0`` : ``JPGFLAG_MATRIX_COLORTRANSFORMATION_NONE``
        | ``1`` : ``JPGFLAG_MATRIX_COLORTRANSFORMATION_YCBCR``
        | ``2`` : ``JPGFLAG_MATRIX_COLORTRANSFORMATION_LSRCT``
        | ``2`` : ``JPGFLAG_MATRIX_COLORTRANSFORMATION_RCT``
        | ``3`` : ``JPGFLAG_MATRIX_COLORTRANSFORMATION_FREEFORM``

    Returns
    -------
    bytes
        The status and any error message of the decoding operation.
    numpy.ndarray or None
        A 1D array of ``np.uint8`` containing the decoded image data. Returns
        ``None`` if the decode fails.
    dict
        A :class:`dict` containing the image parameters.
    """
    # Get the image parameters
    status, param = get_parameters(input_buffer)

    # Pointer to first element in input array
    cdef char *pInput = <char *>np.PyArray_DATA(input_buffer)

    # Create array for output and get pointer to first element
    bpp = ceil(param['precision'] / 8)
    nr_bytes = (
        param['rows'] * param['columns'] * param['nr_components'] * bpp
    )
    output_buffer = np.zeros(nr_bytes, dtype=np.uint8)
    cdef char *pOutput = <char *>np.PyArray_DATA(output_buffer)

    # Decode the data - output is written to output_buffer
    status = Decode(
        pInput,
        pOutput,
        input_buffer.shape[0],
        output_buffer.shape[0],
        colourspace
    )

    return status, output_buffer, param


def get_parameters(np.ndarray[np.uint8_t, ndim=1] input_buffer):
    """Return a :class:`dict` containing the JPEG image parameters.

    Parameters
    ----------
    input_buffer : numpy.ndarray
        A 1D array of ``np.uint8`` containing the encoded JPEG image.

    Returns
    -------
    dict
        A :class:`dict` containing the JPEG image parameters.
    """
    cdef JPEGParameters param
    param.columns = 0
    param.rows = 0
    param.samples_per_pixel = 0
    param.bits_per_sample = 0

    # Pointer to the JPEGParameters object
    cdef JPEGParameters *pParam = &param

    # Pointer to first element in input array
    cdef char *pInput = <char *>np.PyArray_DATA(input_buffer)

    # Decode the data - output is written to output_buffer
    status = GetJPEGParameters(
        pInput,
        input_buffer.shape[0],
        pParam
    )

    parameters = {
        'rows' : param.rows,
        'columns' : param.columns,
        'nr_components' : param.samples_per_pixel,
        'precision' : param.bits_per_sample,
    }

    return status, parameters


cdef extern from "cmd/reconstruct.hpp":
    cdef void Reconstruct(
        const char *inArray,
        const char *outArray,
        int colortrafo,
        const char *alpha,
        bool upsample,
    )


def reconstruct(fin, fout, colourspace, falpha, upsample):
    """Decode the JPEG file in `fin` and write it to `fout` as PNM.

    Parameters
    ----------
    fin : bytes
        The path to the JPEG file to be decoded.
    fout : bytes
        The path to the decoded PNM file.
    colourspace : int
        The colourspace transform to apply.
    falpha : bytes or None
        The path where any decoded alpha channel data will be written,
        otherwise ``None`` to not write alpha channel data. Equivalent to the
        ``-al file`` flag.
    upsample : bool
        ``True`` to disable automatic upsampling, equivalent to the ``-U``
        flag.
    """
    if falpha is None:
        Reconstruct(fin, fout, colourspace, NULL, upsample)
    else:
        Reconstruct(fin, fout, colourspace, falpha, upsample)




import os
import sys
from pathlib import Path
import platform
import setuptools
from setuptools import setup, find_packages
from setuptools.extension import Extension
import subprocess
import distutils.sysconfig

import numpy

PACKAGE_DIR = Path(__file__).parent / "libjpeg"
LIBJPEG_SRC = PACKAGE_DIR / 'src' / 'libjpeg'
INTERFACE_SRC = PACKAGE_DIR / 'src' / 'interface'


def get_mscv_args():
    """Return a list of compiler args for MSVC++'s compiler."""
    flags = [
        '/GS',  # Buffer security check
        '/W3',  # Warning level
        '/Zc:wchar_t',  # Use windows char type
        '/Zc:inline',  # Remove unreferenced function or data (...)
        '/Zc:forScope',
        '/Od',  # Disable optimisation
        '/Oy-',  # (x86 only) don't omit frame pointer
        '/openmp-',  # Disable #pragma omp directive
        '/FC',  # Display full path of source code files
        '/fp:precise',  # Floating-point behaviour
        '/Gd',  # (x86 only) use __cdecl calling convention
        '/GF-',  # Disable string pooling
        '/GR',  # Enable run-time type info
        '/RTC1',  # Enable run-time error checking
        '/MT',  # Create multithreading executable
        # /D defines constants and macros
        '/D_UNICODE',
        '/DUNICODE',
    ]

    # Set the architecture based on system architecture and Python
    is_x64 = platform.architecture()[0] == '64bit'
    if is_x64 and sys.maxsize > 2**32:
        flags.append('/DWIN64=1')
    else:
        # Architecture is 32-bit, or Python is 32-bit
        flags.append('/DWIN32=1')

    return flags


def get_gcc_args():
    """Return a list of compiler and linker args for GCC/clang.

    The args are determined by running the src/libjpeg/configure script then
    parsing src/libjpeg/automakefile for the relevant values.

    Returns
    -------
    dict
        A dict with keys COMPILER_CMD, CC_ONLY, SETTINGS, PREFIX,
        PTHREADCFLAGS, PTHREADLDFLAGS, PTHREADLIBS, HWTYPE, HAVE_ADDONS,
        BITSIZE, ADDOPTS, LIB_OPTS, EXTRA_LIBS, CPU, TUNE.
    """
    # Run configure script once
    # Using GCC or clang, run `configure` bash script once
    if 'config.log' not in os.listdir(LIBJPEG_SRC):
        # Needs to be determined before changing the working dir
        fpath = os.path.abspath(LIBJPEG_SRC)
        # Needs to be run from within the src/libjpeg directory
        current_dir = os.getcwd()
        os.chdir(LIBJPEG_SRC)
        subprocess.call([os.path.join(fpath, 'configure')])
        os.chdir(current_dir)

    # Get compilation options
    with open(os.path.join(LIBJPEG_SRC, 'automakefile')) as fp:
        lines = fp.readlines()

    lines = [ll for ll in lines if not ll.startswith('#')]
    opts = [ll.split('=', 1) for ll in lines]
    opts = {vv[0].strip():list(vv[1].strip().split(' ')) for vv in opts}

    return opts


def get_source_files():
    """Return a list of paths to the source files to be compiled."""
    source_files = [
        PACKAGE_DIR / '_libjpeg.pyx',
        INTERFACE_SRC /'decode.cpp',
        INTERFACE_SRC /'streamhook.cpp',
    ]
    for p in LIBJPEG_SRC.glob('*/*'):
        if p.suffix == '.cpp':
            source_files.append(p)

    # Source files must always be relative to the setup.py directory
    source_files = [p.relative_to(PACKAGE_DIR.parent) for p in source_files]

    return source_files


# Compiler and linker arguments
extra_compile_args = []
extra_link_args = []
if platform.system() == 'Windows':
    os.environ['LIB'] = os.path.abspath(
        os.path.join(sys.executable, '../', 'libs')
    )
    extra_compile_args = get_mscv_args()
elif platform.system() in ['Darwin', 'Linux']:
    # Skip configuration if running with `sdist`
    if 'sdist' not in sys.argv:
        opts = get_gcc_args()
        extra_compile_args += opts['ADDOPTS']
        extra_link_args += opts['EXTRA_LIBS']


extensions = [
    Extension(
        '_libjpeg',
        [os.fspath(p) for p in get_source_files()],
        language='c++',
        include_dirs=[
            LIBJPEG_SRC,
            INTERFACE_SRC,
            numpy.get_include(),
            distutils.sysconfig.get_python_inc(),
        ],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
]

with open(PACKAGE_DIR / '_version.py') as f:
    exec(f.read())

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name = 'pylibjpeg-libjpeg',
    description = (
        "A Python wrapper for libjpeg, with a focus on use as a plugin for "
        "for pylibjpeg"
    ),
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    version = __version__,
    author = "scaramallion",
    author_email = "scaramallion@users.noreply.github.com",
    url = "https://github.com/pydicom/pylibjpeg-libjpeg",
    license = "GPL V3.0",
    keywords = (
        "dicom pydicom python jpg jpeg jpg-ls jpeg-ls libjpeg pylibjpeg"
    ),
    classifiers = [
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Intended Audience :: Developers",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Science/Research",
        "Development Status :: 5 - Production/Stable",
        "Natural Language :: English",
        "Programming Language :: C++",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
        "Topic :: Software Development :: Libraries",
    ],
    packages = find_packages(),
    package_data = {'': ['*.txt', '*.cpp', '*.h', '*.hpp', '*.pyx']},
    include_package_data = True,
    zip_safe = False,
    python_requires = ">=3.7",
    install_requires = [
        "numpy >= 1.20; python_version == '3.7'",
        "numpy >= 1.22; python_version >= '3.8'",
    ],
    ext_modules = extensions,
    # Plugin registrations
    entry_points={
        'pylibjpeg.pixel_data_decoders': [
            "1.2.840.10008.1.2.4.50 = libjpeg:decode_pixel_data",
            "1.2.840.10008.1.2.4.51 = libjpeg:decode_pixel_data",
            "1.2.840.10008.1.2.4.57 = libjpeg:decode_pixel_data",
            "1.2.840.10008.1.2.4.70 = libjpeg:decode_pixel_data",
            "1.2.840.10008.1.2.4.80 = libjpeg:decode_pixel_data",
            "1.2.840.10008.1.2.4.81 = libjpeg:decode_pixel_data",
        ],
        'pylibjpeg.jpeg_decoders': 'libjpeg = libjpeg:decode',
        'pylibjpeg.jpeg_ls_decoders': 'libjpeg = libjpeg:decode',
        'pylibjpeg.jpeg_xt_decoders': 'libjpeg = libjpeg:decode',
    },
)
